import argparse
import os
import pathlib
import re
from glob import glob
from typing import Optional

import matplotlib.pyplot as plt


def process_file(file_path: str) -> Optional[float]:
    """
    Extract all process-averaged run times, remove the largest value, and compute the average.

    Parameters
    ----------
    file_path : str
        Path to the .out file.

    Returns
    -------
    Optional[float]
        The average of the run times after dropping the largest value, or None if there
        are not enough values.
    """
    runtime_pattern = r"Process-averaged run time:\s+(\d+\.\d+)"

    with open(file_path, "r") as file:
        # Extract all run time values as floats.
        times = [float(match) for match in re.findall(runtime_pattern, file.read())]

        # If fewer than two values are found, return None (not enough data to drop the highest value)
        if len(times) <= 1:
            return

        # Drop the largest value.
        times.remove(max(times))

        # Calculate and return the average of the remaining values.
        return sum(times) / len(times)


def extract_node_count(filename: str) -> Optional[int]:
    """
    Extract the node count from a filename.

    Parameters
    ----------
    filename : str
        The name of the .out file.

    Returns
    -------
    Optional[int]
        The extracted node count as an integer, or None if not found.
    """
    filename_pattern = r"nodes_(\d+)\.out"
    match = re.search(filename_pattern, filename)
    return int(match.group(1)) if match else None


def process_out_files(folder_path: str) -> dict[int, Optional[float]]:
    """
    Processes all .out files in a folder to calculate average run times by node count.

    Parameters
    ----------
    folder_path : str
        Path to the folder containing .out files.

    Returns
    -------
    dict[int, Optional[float]]
        Dictionary mapping each node count to the average run time (or None if insufficient data).
    """
    results = {}
    for file_path in glob(os.path.join(folder_path, "*.out")):
        filename = os.path.basename(file_path)
        node_count = extract_node_count(filename)

        if node_count is not None:
            avg_time = process_file(file_path)
            if avg_time is not None:
                results[node_count] = avg_time

    # Display results
    for nodes, avg_time in sorted(results.items()):
        print(f"Nodes {nodes}: Average time (without max value) = {avg_time:.5f} s")

    return results


def plot_results(results: dict[int, float], scaling_type: str, file_path: pathlib.Path | str) -> None:
    """
    Plot runtime, speedup, and efficiency based on the results dictionary.

    Parameters
    ----------
    results : dict[int, float]
        Dictionary containing average run time for each node count.
    scaling_type : str
        The type of scaling considered. Must be either "strong" or "weak".
    """
    # Sort results by number of nodes.
    nodes = sorted(results.keys())
    runtimes = [results[n] for n in nodes]

    if scaling_type == "strong":
        # Calculate speedup (using the runtime with the fewest nodes as the baseline)
        baseline_runtime = runtimes[0]
        derived_quantity = [baseline_runtime / runtime for runtime in runtimes]
        derived_label = "Speedup"
        ideal = nodes
    elif scaling_type == "weak":
        baseline_runtime = runtimes[0]
        derived_quantity = [baseline_runtime / runtime for runtime in runtimes]
        derived_label = "Efficiency"
        ideal = len(nodes) * [1]
        y_lim = [0, 1.1]
    else:
        raise ValueError(f"Unknown scaling type: {scaling_type}")
    print(derived_quantity)

    # Settings
    labelsize = "x-large"

    # Plotting
    fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(12, 5))

    # Runtime vs Number of Nodes
    ax1.set_facecolor("lightgrey")
    ax1.plot(nodes, runtimes, marker="X", ms=10, color="#009682")
    ax1.set_xlabel("Number of nodes", fontweight="bold", fontsize=labelsize)
    ax1.set_ylabel("Runtime / s", fontweight="bold", fontsize=labelsize)
    ax1.grid(True)
    ax1.grid(color="white", linewidth=1.7)
    ax1.tick_params(axis="both", labelsize=labelsize)
    ax1.set_xticks(nodes, [str(node) for node in nodes])

    # Speedup / Efficiency
    ax2.set_facecolor("lightgrey")
    ax2.plot(nodes, ideal, ls="--", color="k", label="Ideal")
    ax2.plot(nodes, derived_quantity, marker="X", ms=10, color="#009682", label="Actual")
    ax2.set_xlabel("Number of nodes", fontweight="bold", fontsize=labelsize)
    ax2.set_ylabel(derived_label, fontweight="bold", fontsize=labelsize)
    ax2.legend()
    ax2.grid(True)
    ax2.grid(color="white", linewidth=1.7)
    ax2.tick_params(axis="both", labelsize=labelsize)
    ax2.set_xticks(nodes, [str(node) for node in nodes])
    if scaling_type == "weak":
        ax2.set_ylim(y_lim)

    plt.tight_layout()
    plt.savefig(pathlib.Path(file_path) / f"{file_path}.pdf")
    plt.show()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        prog="Plot cdist scaling results",
    )
    parser.add_argument(  # Function to optimize
        "--path",
        type=pathlib.Path,
        default="./",
    )
    config = parser.parse_args()
    # Directory containing the .out files
    results = process_out_files(config.path)
    # Plot the results
    plot_results(results, scaling_type=str(config.path).split("_")[0], file_path=pathlib.Path(config.path))
